#  Import modules from Python Standard Library
import datetime
import logging
import Queue
import sys

#  Import third-party Python modules
import serial


####  Site-specific configuration information  ####

InsteonSensor = {
    '00-00-01': ['Broadcast', 'broadcast'],

    #  Sensors in CSB 628
    '13-AF-B9': ['LACE Powerlinc Modem', 'modem'],
    '13-E2-AB': ['Door to Hallway', 'contact'],
    '13-E2-9D': ['Door to Printer', 'contact'],
    '13-E2-F3': ['Fridge Door', 'contact'],
    '13-E2-71': ['Freezer Door', 'contact'],
    
    #  Sensors in Craig's office
    '13-AC-CB': ['Craig\'s Powerlinc Modem', 'modem'],
    '11-BC-A6': ['Craig\'s contact switch #1', 'contact'],
    
#    'XX-YY-ZZ': ['WallOne', 'infrared'],
}

####  /Site-specific configuration information  ####



#  Configure logging
insteon_logger = logging.getLogger('insteon_logger')
insteon_logger.setLevel(logging.DEBUG)


#  Define INSTEON command set
insteon_command = {}

insteon_command['\x50'] = { 
    'expected_bytes':  9, 
    'description':    "INSTEON Standard Message Received",
    'ack_expected':   True
}
insteon_command['\x51'] = { 
    'expected_bytes': 23, 
    'description':    "INSTEON Extended Message Received",
    'ack_expected':   True
}
insteon_command['\x52'] = { 
    'expected_bytes':  2, 
    'description':    "X10 Received",
    'ack_expected':   False
}
insteon_command['\x53'] = { 
    'expected_bytes':  8, 
    'description':    "ALL-Linking Completed",
    'ack_expected':   False
}
insteon_command['\x54'] = {
	'expected_bytes':  1,
	'description':    "Button Event Report",
    'ack_expected':   False
}
insteon_command['\x55'] = {
	'expected_bytes':  0,
	'description' : "User Reset Detected",
    'ack_expected':   False
}
insteon_command['\x56'] = {
	'expected_bytes':  4,
	'description':    "ALL-Link Cleanup Failure Report",
    'ack_expected':   False
}
insteon_command['\x57'] = {
	'expected_bytes':  8,
	'description':    "ALL-Link Record Response",
    'ack_expected':   False
}
insteon_command['\x58'] = {
	'expected_bytes':  1,
	'description':    "ALL-Link Cleanup Status Report",
    'ack_expected':   True
}
insteon_command['\x60'] = {
	'expected_bytes':  7,
	'description':    "Get IM Info",
    'ack_expected':   True
}
insteon_command['\x61'] = {
	'expected_bytes':  3,
	'description':    "Send ALL-Link Command",
    'ack_expected':   True
}
insteon_command['\x62'] = {
	'expected_bytes':  7,
	'description':    "Send INSTEON Standard or Extended Message",
    'ack_expected':   True
}
insteon_command['\x63'] = {
	'expected_bytes':  3,
	'description':    "Send X10",
    'ack_expected':   True
}
insteon_command['\x64'] = {
	'expected_bytes':  3,
	'description':    "Start ALL-Linking",
    'ack_expected':   True
}
insteon_command['\x65'] = {
	'expected_bytes':  1,
	'description':    "Cancel ALL-Linking",
    'ack_expected':   True
}
insteon_command['\x66'] = {
	'expected_bytes':  4,
	'description':    "Set Host Device Category",
    'ack_expected':   True
}
insteon_command['\x67'] = {
	'expected_bytes':  1,
	'description':    "Reset the IM",
    'ack_expected':   True
}
insteon_command['\x68'] = {
	'expected_bytes':  2,
	'description':    "Set INSTEON ACK Message Byte",
    'ack_expected':   True
}
insteon_command['\x69'] = {
	'expected_bytes':  1,
	'description':    "Get First ALL-Link Record",
    'ack_expected':   True
}
insteon_command['\x6A'] = {
	'expected_bytes':  1,
	'description':    "Get Next ALL-Link Record",
    'ack_expected':   True
}
insteon_command['\x6B'] = {
	'expected_bytes':  2,
	'description':    "Set IM Configuration",
    'ack_expected':   True
}
insteon_command['\x6C'] = {
	'expected_bytes':  1,
	'description':    "Get ALL-Link Record for Sender",
    'ack_expected':   True
}
insteon_command['\x6D'] = {
	'expected_bytes':  1,
	'description':    "LED On",
    'ack_expected':   True
}
insteon_command['\x6E'] = {
	'expected_bytes':  1,
	'description':    "LED Off",
    'ack_expected':   True
}
insteon_command['\x6F'] = {
	'expected_bytes': 10,
	'description':    "Manage ALL-Link Record",
    'ack_expected':   True
}
insteon_command['\x70'] = {
	'expected_bytes':  2,
	'description':    "Set INSTEON NAK Message Byte",
    'ack_expected':   True
}
insteon_command['\x71'] = {
	'expected_bytes':  3,
	'description':    "Set INSTEON ACK Message Two Bytes",
    'ack_expected':   True
}
insteon_command['\x72'] = {
	'expected_bytes':  1,
	'description':    "RF Sleep",
    'ack_expected':   True
}
insteon_command['\x73'] = {
	'expected_bytes':  4,
	'description':    "Get IM Configuration",
    'ack_expected':   True
}


insteon_command_queue = Queue.Queue()


def bitstring(s):
    return str(s) if s<=1 else bitstring(s>>1) + str(s&1)


def get_all_link_record_response(serial_port):
    (command_byte, insteon_data) = read_insteon_command_blocking(serial_port)
    print "\n  INSTEON Command: %s" % insteon_command[command_byte]['description']

    if command_byte == '\x57':
        print_all_link_record_response(insteon_data)
        ping_insteon_device(serial_port, insteon_data[2:5])
    else:
        print "ERROR: Expected ALL-Link Record Response, received %d" % ord(command_byte)
        # TODO: Raise exception


def process_standard_insteon_command(insteon_data):
    timestamp = datetime.datetime.now()

    from_address = hex_string_for_insteon_address(insteon_data[0:3])
    to_address   = hex_string_for_insteon_address(insteon_data[3:6])

    from_sensor_name = InsteonSensor[from_address][0]
    to_sensor_name   = InsteonSensor[to_address][0]

    from_sensor_type = InsteonSensor[from_address][1]
    to_sensor_type   = InsteonSensor[to_address][1]


    print "\nReceived INSTEON Standard message"
    print "FROM:  %s  (%s)" % (from_sensor_name, from_address)
    print "TO:    %s  (%s)" % (to_sensor_name,   to_address)

    if str(from_sensor_type) == 'contact':
        contact_switch_status = ord(insteon_data[7])
        if contact_switch_status == 19:
            message = timestamp.isoformat() + ', ' + from_address + ', ' + from_sensor_name + ', ' + from_sensor_type + ',  Closed'
            insteon_logger.info(message)
            print message
        elif contact_switch_status == 17:
            message = timestamp.isoformat() + ', ' + from_address + ', ' + from_sensor_name + ', ' + from_sensor_type + ',  Open'
            insteon_logger.info(message)
            print message
        else:
            print "Unknown status code for contact switch '%s': %d" % (from_sensor.name, contact_switch_status)
    elif str(from_sensor_type) == 'infrared':
        message = timestamp.isoformat() + ', ' + from_address + ', ' + from_sensor_name + ', ' + from_sensor_type + ',  Triggered'
        insteon_logger.info(messge)
        print message


def hex_string_for_insteon_address(address):
    return "%.2X-%.2X-%.2X" % (ord(address[0]), ord(address[1]), ord(address[2]))


#  TODO: Implement singleton design pattern
def insteon_reader(portname, logfile_prefix):
    insteon_logfile = logging.FileHandler(logfile_prefix + '_insteon.log')
    insteon_logfile.setLevel(logging.DEBUG)
    insteon_logger.addHandler(insteon_logfile)

    try:
        print "Connecting to INSTEON interface on port: %s\n" % portname
        serial_port = serial.Serial(port=portname, baudrate=19200, timeout=0.125)
    except serial.SerialException:
        print "ERROR: Unable to open serial port"
        sys.exit(0)

    try:
        print "Searching for INSTEON devices..."
        print_linked_insteon_devices(serial_port)
        print "\nListening for INSTEON messages..."

        while 1:
            try:
                command = insteon_command_queue.get_nowait()
                print_linked_insteon_devices(serial_port)
            except Queue.Empty:
                (rvalue, command_byte, insteon_data) = read_insteon_command_nonblocking(serial_port)

                if rvalue:
                    if command_byte == '\x50':
                        process_standard_insteon_command(insteon_data)
                    else:
                        print "Received insteon command: %s" % insteon_command[command_byte]['description']

    except serial.SerialException:
        print "ERROR: Unable to read data from port '%s'" % portname


def is_valid_insteon_command(command_byte):
    if command_byte in insteon_command:
        return True
    else:
        return False


def ping_insteon_device(serial_port, insteon_id_bytes):
    insteon_id_string = hex_string_for_insteon_address(insteon_id_bytes)

    print "      Pinging INSTEON device '{}' ({})...".format(insteon_id_string, InsteonSensor[insteon_id_string][0])
    message_flags = '\x0F'
    ping_command = '\x02\x62' + insteon_id_bytes + message_flags + '\x0F\x00'

    serial_port.write(ping_command)

    (command_byte, insteon_data) = read_insteon_command_blocking(serial_port)
    if command_byte == '\x62':
        if was_insteon_ack_received(insteon_data):
            if insteon_id_bytes == insteon_data[0:3]:
                if (ord(insteon_data[3]) & 16) == 16:
                    print '      Device responded to ping.'
                else:
                    print '      Device DID NOT respond to ping.'
            else:
                print "      Received response from '%s' instead of '%s'" % (hex_string_for_insteon_address(insteon_data[0:3]), insteon_id_string)
        else:
            print '      Received a NAK in reponse to ping request.'
    else:
        print '      Did not receive INSTEON Standard Message, received INSTEON Command: %s' % insteon_command[command_byte]['description']


def print_all_link_record_response(data):
    insteon_id = hex_string_for_insteon_address(data[2:5])

    print "    ALL-Link Record Flags: %s" % bitstring(ord(data[0])) # ord(data[0])
    print "    ALL-Link Group:        %d" % ord(data[1])
    sensor = InsteonSensor[insteon_id]
    print "    Link Data 1:           %d" % ord(data[5])
    print "    Link Data 2:           %d" % ord(data[6])
    print "    Link Data 3:           %d" % ord(data[7])


def print_linked_insteon_devices(serial_port):
    if request_first_all_link_record(serial_port):
        get_all_link_record_response(serial_port)

        while request_next_all_link_record(serial_port):
            get_all_link_record_response(serial_port)
        else:
            print "No more INSTEON devices found"
    else:
        print "ERROR: No devices found"


def put_insteon_command_in_queue(ic):
    insteon_command_queue.put(ic)


def read_data_for_insteon_command(command_byte, serial_port):
    length = insteon_command[command_byte]['expected_bytes']

    if length != 0:
        data = serial_port.read(length)
        if length != len(data):
            print "WARNING: Expected %d bytes, but received %d bytes" % (length, len(data))
        return data
    else:
        return ''


def read_insteon_command_blocking(serial_port):
    read_until_stx_blocking(serial_port)
    command_byte = read_until_insteon_command_blocking(serial_port)
    insteon_data = read_data_for_insteon_command(command_byte, serial_port)
    return (command_byte, insteon_data)


def read_insteon_command_nonblocking(serial_port):
    if read_until_stx_nonblocking(serial_port):
        command_byte = read_until_insteon_command_blocking(serial_port)
        insteon_data = read_data_for_insteon_command(command_byte, serial_port)
        return (True, command_byte, insteon_data)
    else:
        return (False, False, False)


def read_until_insteon_command_blocking(serial_port):
    while 1:
        data = serial_port.read(1)
        if len(data) > 0:
            if is_valid_insteon_command(data[0]):
                return data[0]
            else:
                print "WARNING: Expected INSTEON command, but received: %d" % ord(data[0])
                read_until_stx(serial_port)


def read_until_stx_blocking(serial_port):
    while 1:
        data = serial_port.read(1)
        if len(data) > 0:
            if data[0] == '\x02':
                return
            else:
                print "WARNING: Searching for STX command, ignoring INSTEON data: %d" % ord(data[0])


def read_until_stx_nonblocking(serial_port):
    data = serial_port.read(1)
    if len(data) > 0:
        if data[0] == '\x02':
            return True
        else:
            print "WARNING: Searching for STX command, ignoring INSTEON data: %d" % ord(data[0])
            print "   INFO: Making a blocking call searching for STX command..."
            read_until_stx_blocking(serial_port)
    else:
        return False


def request_first_all_link_record(serial_port):
    serial_port.write('\x02\x69')
    (command_byte, insteon_data) = read_insteon_command_blocking(serial_port)
    print "  INSTEON Command: %s" % insteon_command[command_byte]['description']
    return was_insteon_ack_received(insteon_data)


def request_next_all_link_record(serial_port):
    serial_port.write('\x02\x6A')
    (command_byte, insteon_data) = read_insteon_command_blocking(serial_port)
    print "  INSTEON Command: %s" % insteon_command[command_byte]['description']
    return was_insteon_ack_received(insteon_data)


def was_insteon_ack_received(data):
    last_byte = len(data) - 1

    if data[last_byte] == '\x06':
        return True
    elif data[last_byte] == '\x15':
        return False
    else:
        print "WARNING: Unknown response received: %d" % ord(data[last_byte])
        # TODO: Throw exception
